import sys
import os
import numpy as np
import pandas as pd
import bisect

def get_nearest_frame(timepoint, timestamps):
    '''
    Get the nearest frame timestamp for any time point

    This is kinda not true. This returns the index at which you would
    insert the timepoint to retain the sort order of the list, so if you
    use the index on the list of timestamps you will always get the smallest 
    timestamps that is larger than your input timestamp (not alway the closest)

    Args:
        timepoint (float): The timepoint you want a frame time for
        timestamps (list or np.array) The timestamps of each frame

    Returns:
        nearest_frame (int): The index of the next frame in time
    '''
    nearest_frame = bisect.bisect_left(timestamps, timepoint)
    return nearest_frame

def get_trace_around_timepoint(trace, timepoint, timestamps,
                               window_around_timepoint_seconds, frame_rate):
    '''
    Return the values around a timepoint using a window defined in seconds

    Args:
        trace (np.array): The trace values
        timepoint (float): The timepoint around which to apply the window
        timestamps (np.array): Timestamp in seconds for each point in the trace
        window_around_timepoint_seconds (list with len==2):
            [-2, 3] for a window that starts 2 seconds before the timepoint and
            ends 3 seconds after. 
        frame_rate (float): The frame rate at which the trace is collected.
    '''

    assert trace.shape == timestamps.shape

    window = window_around_timepoint_seconds
    frame_for_timepoint = get_nearest_frame(timepoint, timestamps)
    lower_frame = frame_for_timepoint + int((window[0] * frame_rate))
    upper_frame = frame_for_timepoint + int((window[1] * frame_rate))
    trace = np.array(trace[lower_frame:upper_frame])
    timepoints = np.array(timestamps[lower_frame:upper_frame])
    return trace, timepoints

def get_mean_in_window(trace, window_after_trace_start_seconds, frame_rate):
    window = window_after_trace_start_seconds.copy()
    mean = np.nanmean(trace[int(window[0] * frame_rate): int(window[1] * frame_rate)]) 
    return mean

if __name__=="__main__":
    trace = np.arange(100, dtype=float)
    timestamps = np.arange(100, dtype=float)

    a = get_nearest_frame(49.9, timestamps)
    b = get_nearest_frame(50.1, timestamps)

    assert timestamps[a] == 50.0
    assert timestamps[b] == 51.0

    window_around_timepoint_seconds = [-5, 5]
    t_vals, t_ts = get_trace_around_timepoint(trace, 49.9, timestamps,
                                              window_around_timepoint_seconds,
                                              frame_rate=1)

    assert np.all(t_vals == np.array([45., 46., 47., 48., 49., 50., 51., 52., 53., 54.]))

#  def traces_around_timepoints(trace_values, trace_timestamps, event_times, window):
#      '''
#      Get peri-event slices of a trace.
#  
#      Args:
#          trace_values (1d np.array): Trace for one cell
#          trace_timestamps (1d np.array): Timestamps for each trace value
#          event_times (np.array): The times of events you want traces for
#          window (2-tuple): Time range around event times
#  
#      Returns
#          eventlocked_traces (np.array with shape (n_events, n_samples_in_window))
#      '''







