from typing import Tuple

import numpy as np

from scipy.interpolate import RegularGridInterpolator, griddata


def regular_grid_extractor_factory(timestamps: np.ndarray,
                                   lfp_raw: np.ndarray,
                                   channel: int,
                                   method: str = 'linear') -> np.ndarray:
    '''Builds an LFP data extractor using interpolation on a regular grid

    Ignores timestamps less than zero (which result from unaligned
    data segments)

    Parameters
    ----------
    timestamps : numpy.ndarray
        Associates LFP sample indices with times in seconds.
    lfp_raw : numpy.ndarray
        Dimensions are samples X channels.
    channel : int
        Index of channel to interpolate to regular grid.
    method : str, optional
        Interpolation method ['linear', 'cubic', 'nearest'],
        by default 'linear'.

    Returns
    -------
    numpy.ndarray
        LFP data that has been interpolated to a regular grid.
    '''

    valid_timestamps = (timestamps >= 0)

    return RegularGridInterpolator((timestamps[valid_timestamps],),
                                   lfp_raw[valid_timestamps, channel],
                                   method=method,
                                   bounds_error=False,
                                   fill_value=np.nan)


def make_actual_channel_locations(min_chan: int = 0,
                                  max_chan: int = 384) -> np.ndarray:
    '''Generate x/y locations of Neuropixels recording sites.

          0  8 16 24 32 40 48
    60    *  -  -  -  *  -  -
    50    -  -  -  -  -  -  -
    40    -  -  *  -  -  -  * <-- actual recording site (*)
    30    -  -  -  -  -  -  -
    20    *  -  -  -  *  -  -
    10    -  -  -  -  -  -  -
     0    -  -  *  -  -  -  *

    Parameters
    ----------
    min_chan : int, optional
        Lowest channel number to use, by default 0
    max_chan : int, optional
        Highest channel number to use, by default 384

    Returns
    -------
    actual_channel_locations: numpy.ndarray
        column 1 = x positions in microns
        column 2 = y positions in microns
    '''

    actual_channel_locations = np.zeros((max_chan, 2))

    x_locations = [16, 48, 0, 32]

    for ch in range(min_chan, max_chan):
        actual_channel_locations[ch, 0] = x_locations[ch % 4]
        actual_channel_locations[ch, 1] = np.floor(ch / 2) * 20

    return actual_channel_locations[min_chan:, :]


def make_interp_channel_locations(min_chan: int = 0,
                                  max_chan: int = 384) -> np.ndarray:
    '''Generate x/y locations for interpolated Neuropixels recording sites.

    This version just returns the central column of interpolated sites.

          0  8 16 24 32 40 48
    60    *  -  -  o  *  -  -
    50    -  -  -  o  -  -  -
    40    -  -  *  o  -  -  * <-- actual recording site (*)
    30    -  -  -  o  -  -  -
    20    *  -  -  o  *  -  -
    10    -  -  -  o  -  -  -
     0    -  -  *  o  -  -  *
                   ^
                   interpolated column sites (o)

    Parameters
    ----------
    min_chan : int, optional
        Lowest channel number to use, by default 0
    max_chan : int, optional
        Highest channel number to use, by default 384

    Returns
    -------
    interp_channel_locations: numpy.ndarray
        column 1 = interpolated x positions in microns
        column 2 = y positions in microns
    '''

    interp_channel_locations = np.zeros((max_chan, 2))

    for ch in range(min_chan, max_chan):
        interp_channel_locations[ch, 0] = 24
        interp_channel_locations[ch, 1] = ch * 10

    return interp_channel_locations[min_chan:, :]


def interp_channel_locs(lfp: np.ndarray,
                        actual_locs: np.ndarray,
                        interp_locs: np.ndarray,
                        method: str = 'cubic') -> Tuple[np.ndarray, float]:
    '''Interpolates single-trial lfp channel locations to account for
    channel stagger.

    Parameters
    ----------
    lfp : numpy.ndarray
        LFP data in the form of: trials x channels x time samples
    actual_locs: numpy.ndarray
        An array of actual x, y locations for all channels in lfp. The
        number of actual_locs should equal the number of channels in the 'lfp'.
    interp_locs: numpy.ndarray
        An array of virtual x, y locations for where channels in lfp
        should be interpolated to.

    method : str, optional
        Interpolation method ['cubic', 'linear', 'nearest'], by default 'cubic'

    Returns
    -------
    Tuple[interp_lfp, spacing]
        interp_lfp: numpy.ndarray
            Channel location interpolated lfp data in the form of:
            trials x channels x time samples
        spacing: float
            Distance between new interpolated virtual channel sites
            (in millimeters)
    '''

    if lfp.shape[1] != actual_locs.shape[0]:
        e_msg = (f"Number of 'lfp' channels ({lfp.shape[1]}) does not "
                 f"match number of 'actual_locs' ({actual_locs.shape[0]})!")
        raise RuntimeError(e_msg)

    spacing = np.mean(np.diff(interp_locs[:, 1])) / 1000

    interp_lfp = np.zeros((lfp.shape[0],  # number of interp trials
                           interp_locs.shape[0],  # number of interp channels
                           lfp.shape[2]))  # number of interp samples

    for trial in range(lfp.shape[0]):  # trials
        trial_data = lfp[trial, :, :]
        for t in range(0, lfp.shape[2]):  # time samples
            interp_lfp[trial, :, t] = griddata(points=actual_locs,
                                               values=trial_data[:, t],
                                               xi=interp_locs,
                                               method=method,
                                               fill_value=0,
                                               rescale=False)

    return (interp_lfp, spacing)
