# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2019. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from argschema import ArgSchema, ArgSchemaParser
from argschema.schemas import DefaultSchema
from argschema.fields import Nested, InputDir, String, Boolean, Float, Dict, Int, NumpyArray


class ProbeInputParameters(DefaultSchema):
    name = String(required=True, help='Identifier for this probe')
    lfp_input_file_path = String(required=True, description="path to original LFP .dat file")
    lfp_timestamps_input_path = String(required=True, description="path to LFP timestamps")
    lfp_data_path = String(required=True, help="Path to LFP data continuous file")
    lfp_timestamps_path = String(required=True, help="Path to LFP timestamps aligned to master clock")
    lfp_channel_info_path = String(required=True, help="Path to LFP channel info")
    total_channels = Int(default=384, help='Total channel count for this probe.')
    surface_channel = Int(required=True, help="Probe surface channel")
    reference_channels = NumpyArray(required=False, help="Probe reference channels")
    lfp_sampling_rate = Float(required=True, help="Sampling rate of LFP data")
    noisy_channels = NumpyArray(required=False, help="Noisy channels to remove")


class LfpSubsamplingParameters(DefaultSchema):
    temporal_subsampling_factor = Int(default=2, description="Ratio of input samples to output samples in time")
    channel_stride = Int(default=4, description="Distance between channels to keep")
    surface_padding = Int(default=40, description="Number of channels above surface to include")
    start_channel_offset = Int(default=2, description="Offset of first channel (from bottom of the probe)")
    reorder_channels = Boolean(default=True, description="Implement channel reordering")
    cutoff_frequency = Float(default=0.1, description="Cutoff frequency for DC offset filter (Butterworth)")
    filter_order = Int(default=1, description="Order of DC offset filter (Butterworth)")
    remove_reference_channels = Boolean(default=False,
                                        description="indicates whether references should be removed from output")
    remove_channels_out_of_brain = Boolean(default=False,
                                           description="indicates whether to remove channels outside the brain")
    remove_noisy_channels = Boolean(default=False,
                                    description="indicates whether noisy channels should be removed from output")


class InputParameters(ArgSchema):
    probes = Nested(ProbeInputParameters, many=True, help='Probes for LFP subsampling')
    lfp_subsampling = Nested(LfpSubsamplingParameters, help='Parameters for this module')


class OutputSchema(DefaultSchema):
    input_parameters = Nested(InputParameters, description="Input parameters the module was run with", required=True)


class ProbeOutputParameters(DefaultSchema):
    name = String(required=True, help='Identifier for this probe.')
    lfp_data_path = String(required=True, help='Output subsampled data file.')
    lfp_timestamps_path = String(required=True, help='Timestamps for subsampled data.')
    lfp_channel_info_path = String(required=True, help='LFP channels from that was subsampled.')


class OutputParameters(OutputSchema):
    probe_outputs = Nested(ProbeOutputParameters, many=True, required=True, help='probewise outputs')
