import argparse
import os
import subprocess
import contextlib
import shutil


CURR_FILE_DIR = os.path.dirname(os.path.abspath(__file__))
DOCKERFILE_STAGE_1 = os.path.join(CURR_FILE_DIR, 'stage_1', 'Dockerfile')
DOCKERFILE_STAGE_2 = os.path.join(CURR_FILE_DIR, 'stage_2', 'Dockerfile')
DOCKERFILE_STAGE_3 = os.path.join(CURR_FILE_DIR, 'stage_3', 'Dockerfile')
DOCKERFILE_STAGE_4 = os.path.join(CURR_FILE_DIR, 'stage_4', 'Dockerfile')
MODELFILE_CACHE_LOC = os.path.join(CURR_FILE_DIR, 'stage_1', 'modelfile.zip')


def run_rule(rule, **kwargs):

    if rule == 'clean':

        with contextlib.suppress(FileNotFoundError):
            os.remove(MODELFILE_CACHE_LOC)

    elif rule == 'build:stage-1':

        # Download and cache modelfile:
        modelfile = kwargs.get('modelfile')
        if not modelfile:

            if not os.path.exists(MODELFILE_CACHE_LOC):
                from google.cloud import storage

                source_bucketname = 'dlc-eye-tracking-models'
                source_filename = 'universal_eye_tracking-peterl-2019-07-10.zip'
                target_filename = MODELFILE_CACHE_LOC

                client = storage.Client()
                bucket = client.get_bucket(source_bucketname)
                blob = bucket.blob(source_filename)
                blob.download_to_filename(target_filename)
            modelfile = MODELFILE_CACHE_LOC
        assert os.path.exists(modelfile)

        subprocess.check_call(['docker', 'build',
                               '--build-arg', 'MODELFILE={modelfile}'.format(modelfile='modelfile.zip'),
                               '-t', 'dlc-eye-tracking:stage-1', '-f', DOCKERFILE_STAGE_1, 'stage_1'])

    elif rule == 'build:stage-2':
        subprocess.check_call(['docker', 'build',
                               '-t', 'dlc-eye-tracking:stage-2', '-f', DOCKERFILE_STAGE_2, 'stage_2'])

    elif rule == 'build:stage-4':
        subprocess.check_call(['docker', 'build',
                               '-t', 'dlc-eye-tracking:stage-4', '-f', DOCKERFILE_STAGE_4, 'stage_4'])

    elif rule == 'build:stage-3':

        # Download and cache modelfile:
        modelfile = kwargs.get('modelfile')
        if not modelfile:

            if not os.path.exists(MODELFILE_CACHE_LOC):
                from google.cloud import storage

                source_bucketname = 'dlc-eye-tracking-models'
                source_filename = 'universal_eye_tracking-peterl-2019-07-10.zip'
                target_filename = MODELFILE_CACHE_LOC

                client = storage.Client()
                bucket = client.get_bucket(source_bucketname)
                blob = bucket.blob(source_filename)
                blob.download_to_filename(target_filename)
            modelfile = MODELFILE_CACHE_LOC
        assert os.path.exists(modelfile)
        shutil.copyfile(modelfile, os.path.join(CURR_FILE_DIR, 'stage_3', 'modelfile.zip'))

        subprocess.check_call(['docker', 'build',
                               '--build-arg', 'MODELFILE={modelfile}'.format(modelfile='modelfile.zip'),
                               '-t', 'dlc-eye-tracking:stage-3', '-f', DOCKERFILE_STAGE_3, 'stage_3'])

    elif rule in ['run:stage-1', 'run:stage-2']:
        assert kwargs['modelfile'] is None
        video_input_file = kwargs.get('video_input_file')
        stage = rule.split(':')[-1]

        subprocess.check_call(['docker', 'run',
                               '--runtime=nvidia',
                               '-e', 'VIDEO_INPUT_FILE={}'.format(video_input_file), 'dlc-eye-tracking:{}'.format(stage)])

    elif rule in ['tag:stage-1', 'tag:stage-2', 'tag:stage-3', 'tag:stage-4']:
        stage = rule.split(':')[-1]
        subprocess.check_call(['docker', 'tag', 'dlc-eye-tracking:{}'.format(stage), 'us.gcr.io/aibs-pilot/dlc-eye-tracking:{}'.format(stage)])

    elif rule in ['tag-aibs:stage-1', 'tag-aibs:stage-2', 'tag-aibs:stage-3']:
        stage = rule.split(':')[-1]
        subprocess.check_call(['docker', 'tag', 'dlc-eye-tracking:{}'.format(stage), 'docker.aibs-artifactory.corp.alleninstitute.org/dlc-eye-tracking:{}'.format(stage)])

    elif rule in ['push:stage-1', 'push:stage-2', 'push:stage-3', 'push:stage-4']:
        stage = rule.split(':')[-1]
        subprocess.check_call(['docker', 'push', 'us.gcr.io/aibs-pilot/dlc-eye-tracking:{}'.format(stage)])

    elif rule in ['push-aibs:stage-1', 'push-aibs:stage-2', 'push-aibs:stage-3']:
        stage = rule.split(':')[-1]
        subprocess.check_call(['docker', 'push', 'docker.aibs-artifactory.corp.alleninstitute.org/dlc-eye-tracking:{}'.format(stage)])

    elif rule == 'build:all':
        run_rule('build:stage-1', **kwargs)
        run_rule('build:stage-2', **kwargs)
        run_rule('build:stage-3', **kwargs)
        run_rule('build:stage-4', **kwargs)

    elif rule == 'tag:all':
        run_rule('tag:stage-1', **kwargs)
        run_rule('tag:stage-2', **kwargs)
        run_rule('tag:stage-3', **kwargs)
        run_rule('tag:stage-4', **kwargs)

    elif rule == 'push:all':
        run_rule('push:stage-1', **kwargs)
        run_rule('push:stage-2', **kwargs)
        run_rule('push:stage-3', **kwargs)
        run_rule('push:stage-4', **kwargs)

    elif rule == 'all':
        run_rule('build:all', **kwargs)
        run_rule('tag:all', **kwargs)
        run_rule('push:all', **kwargs)
    else:

        raise RuntimeError('Invalid rule: {}'.format(rule))


if __name__ == "__main__":

    # Sanity check:
    for filename in [DOCKERFILE_STAGE_1, DOCKERFILE_STAGE_2]:
        assert os.path.exists(filename)

    parser = argparse.ArgumentParser()
    parser.add_argument("rule", help="Rule to run", choices=['build:stage-1', 'run:stage-1', 'tag:stage-1', 'push:stage-1',
                                                             'build:stage-2', 'run:stage-2', 'tag:stage-2', 'push:stage-2',
                                                             'build:stage-3', 'tag:stage-3', 'push:stage-3',
                                                             'build:stage-4', 'tag:stage-4', 'push:stage-4',
                                                             'tag-aibs:stage-1', 'tag-aibs:stage-2', 'push-aibs:stage-1', 'push-aibs:stage-2',
                                                             'clean', 'build:all', 'tag:all', 'push:all', 'all'], nargs='+')
    parser.add_argument("--modelfile", help="DLC model zip file location", type=str)
    parser.add_argument("--video_input_file", help="input artifact", type=str)

    args = vars(parser.parse_args())
    for rule in args.pop('rule'):
        run_rule(rule, **args)
