import logging
import pprint
import sys
import argparse
import os

import argschema
import requests

from ._schemas import InputParameters, OutputParameters
from . import cases
from .image_series_gridder import ImageSeriesGridder


def get_inputs_from_lims(host, image_series_id, output_root, job_queue, strategy):
    
    uri = ''.join('''
        {}/input_jsons?
        object_id={}&
        object_class=ImageSeries&
        strategy_class={}&
        job_queue_name={}
    '''.format(host, image_series_id, strategy, job_queue).split())
    response = requests.get(uri)
    data = response.json()

    if len(data) == 1 and 'error' in data:
        raise ValueError('bad request uri: {} ({})'.format(uri, data['error']))

    data['storage_directory'] = os.path.join(output_root, os.path.split(data['storage_directory'])[-1])
    data['grid_prefix'] = os.path.join(output_root, os.path.split(data['grid_prefix'])[-1])
    data['accumulator_prefix'] = os.path.join(output_root, os.path.split(data['accumulator_prefix'])[-1])

    return data


def write_or_print_outputs(data, parser):
    data.update({'input_parameters': parser.args})
    if 'output_json' in parser.args:
        parser.output(data, indent=2)
    else:
        print(parser.get_output_json(data))    


def run_grid(args):

    try:
        case = cases[args['case']]
    except KeyError:
        logging.error('unrecognized case: {}'.format(args['case']))
        raise

    sub_images = args['sub_images']
    

    input_dimensions = [sub_images[0]['dimensions']['column'], 
                        sub_images[0]['dimensions']['row'], 
                        args['sub_image_count']]  

    input_spacing = [sub_images[0]['spacing']['column'], 
                     sub_images[0]['spacing']['row'], 
                     args['image_series_slice_spacing']] 

    for ii, si in enumerate(sub_images):
        del si['dimensions']
        del si['spacing']
        si['polygon_info'] = si['polygons']
        del si['polygons']
    sub_images = sorted(sub_images, key=lambda si: si['specimen_tissue_index'])
    logging.info('{} sub images with indices: {}'.format(
        len(sub_images), [si['specimen_tissue_index'] for si in sub_images])
    )

    output_dimensions = [args['reference_dimensions']['slice'], 
                         args['reference_dimensions']['row'], 
                         args['reference_dimensions']['column']]

    output_spacing = [args['reference_spacing']['slice'], 
                      args['reference_spacing']['row'], 
                      args['reference_spacing']['column']]

    subimage_kwargs = {'cls': case['subimage']}
    if args['filter_bit'] is not None:
        subimage_kwargs['filter_bit'] = args['filter_bit']

    gridder = ImageSeriesGridder(
        in_dims=input_dimensions, 
        in_spacing=input_spacing, 
        out_dims=output_dimensions, 
        out_spacing=output_spacing, 
        reduce_level=args['reduce_level'], 
        subimages=sub_images, 
        subimage_kwargs=subimage_kwargs, 
        nprocesses=args['nprocesses'], 
        affine_params=args['affine_params'], 
        dfmfld_path=args['deformation_field_path']
    )

    gridder.setup_subimages()
    gridder.build_coarse_grids()

    writer = case['writer']
    paths = writer(gridder, args['grid_prefix'], args['accumulator_prefix'], target_spacings=args['target_spacings'])

    return {'output_file_paths': paths}


def main():

    logging.basicConfig(format='%(asctime)s - %(process)s - %(levelname)s - %(message)s')

    # TODO replace with argschema implementation of multisource parser
    remaining_args = sys.argv[1:]
    input_data = {}
    if '--get_inputs_from_lims' in sys.argv:
        lims_parser = argparse.ArgumentParser(add_help=False)
        lims_parser.add_argument('--host', type=str, default='http://lims2')
        lims_parser.add_argument('--job_queue', type=str, default=None)
        lims_parser.add_argument('--strategy', type=str,default= None)
        lims_parser.add_argument('--image_series_id', type=int, default=None)
        lims_parser.add_argument('--output_root', type=str, default= None)

        lims_args, remaining_args = lims_parser.parse_known_args(remaining_args)
        remaining_args = [item for item in remaining_args if item != '--get_inputs_from_lims']
        input_data = get_inputs_from_lims(**lims_args.__dict__)

    parser = argschema.ArgSchemaParser(
        args=remaining_args,
        input_data=input_data,
        schema_type=InputParameters,
        output_schema_type=OutputParameters,
    )

    output = run_grid(parser.args)
    write_or_print_outputs(output, parser)


if __name__ == '__main__':
    main()